@@ -307,11 +307,14 @@ import DataFrame.Operations.Permutation as Permutation (
 import DataFrame.Operations.Statistics as Statistics (
     correlation,
     frequencies,
+    genericPercentile,
     imputeWith,
     interQuartileRange,
     mean,
     meanMaybe,
     median,
+    medianMaybe,
+    percentile,
     skewness,
     standardDeviation,
     sum,
