@@ -14,8 +14,9 @@ module DataFrame.DecisionTree where
 import qualified DataFrame.Functions as F
 import DataFrame.Internal.Column
 import DataFrame.Internal.DataFrame (DataFrame (..), unsafeGetColumn)
-import DataFrame.Internal.Expression (Expr (..), eSize, interpret)
-import DataFrame.Internal.Statistics (percentileOrd')
+import DataFrame.Internal.Expression (Expr (..), eSize)
+import DataFrame.Internal.Interpreter (interpret)
+import DataFrame.Internal.Statistics (percentile', percentileOrd')
 import DataFrame.Internal.Types
 import DataFrame.Operations.Core (columnNames, nRows)
 import DataFrame.Operations.Statistics (percentile)
@@ -27,6 +28,7 @@ import Data.Containers.ListUtils (nubOrd)
 import Data.Function (on)
 import Data.List (foldl', maximumBy, sortBy)
 import qualified Data.Map.Strict as M
+import Data.Maybe
 import qualified Data.Text as T
 import Data.Type.Equality
 import qualified Data.Vector as V
@@ -35,25 +37,33 @@ import Type.Reflection (typeRep)
 
 import DataFrame.Functions ((.<), (.<=), (.==), (.>), (.>=))
 
-data TreeConfig = TreeConfig
+data TreeConfig
+    = TreeConfig
     { maxTreeDepth :: Int
     , minSamplesSplit :: Int
+    , minLeafSize :: Int
     , synthConfig :: SynthConfig
     }
+    deriving (Eq, Show)
 
 data SynthConfig = SynthConfig
     { maxExprDepth :: Int
+    , boolExpansion :: Int
     , percentiles :: [Int]
+    , complexityPenalty :: Double
     , enableStringOps :: Bool
     , enableCrossCols :: Bool
     , enableArithOps :: Bool
     }
+    deriving (Eq, Show)
 
 defaultSynthConfig :: SynthConfig
 defaultSynthConfig =
     SynthConfig
         { maxExprDepth = 2
+        , boolExpansion = 2
         , percentiles = [0, 10 .. 100]
+        , complexityPenalty = 0.05
         , enableStringOps = True
         , enableCrossCols = True
         , enableArithOps = True
@@ -62,8 +72,9 @@ defaultSynthConfig =
 defaultTreeConfig :: TreeConfig
 defaultTreeConfig =
     TreeConfig
-        { maxTreeDepth = 10
+        { maxTreeDepth = 4
         , minSamplesSplit = 5
+        , minLeafSize = 1
         , synthConfig = defaultSynthConfig
         }
 
@@ -98,7 +109,7 @@ buildTree cfg depth target conds df
     | depth <= 0 || nRows df <= minSamplesSplit cfg =
         Lit (majorityValue @a target df)
     | otherwise =
-        case findBestSplit @a target conds df of
+        case findBestSplit @a cfg target conds df of
             Nothing -> Lit (majorityValue @a target df)
             Just bestCond ->
                 let (dfTrue, dfFalse) = partitionDataFrame bestCond df
@@ -208,11 +219,54 @@ generateConditionsOld cfg df =
                     percentiles = map (Lit . (`percentileOrd'` col)) [1, 25, 75, 99]
                  in
                     map (Col @a colName .==) percentiles
-            (OptionalColumn (col :: V.Vector a)) ->
-                let
-                    percentiles = map (Lit . (`percentileOrd'` col)) [1, 25, 75, 99]
-                 in
-                    map (Col @a colName .==) percentiles
+            (OptionalColumn (col :: V.Vector (Maybe a))) -> case sFloating @a of
+                STrue ->
+                    let
+                        doubleCol =
+                            VU.convert
+                                (V.map fromJust (V.filter isJust (V.map (fmap (realToFrac @a @Double)) col)))
+                     in
+                        zipWith
+                            ($)
+                            [ (Col @(Maybe a) colName .==)
+                            , (Col @(Maybe a) colName .<=)
+                            , (Col @(Maybe a) colName .>=)
+                            ]
+                            ( Lit Nothing
+                                : map
+                                    ( Lit
+                                        . Just
+                                        . realToFrac
+                                        . (`percentile'` doubleCol)
+                                    )
+                                    (percentiles cfg)
+                            )
+                SFalse -> case sIntegral @a of
+                    STrue ->
+                        let
+                            doubleCol =
+                                VU.convert
+                                    (V.map fromJust (V.filter isJust (V.map (fmap (fromIntegral @a @Double)) col)))
+                         in
+                            zipWith
+                                ($)
+                                [ (Col @(Maybe a) colName .==)
+                                , (Col @(Maybe a) colName .<=)
+                                , (Col @(Maybe a) colName .>=)
+                                ]
+                                ( Lit Nothing
+                                    : map
+                                        ( Lit
+                                            . Just
+                                            . round
+                                            . (`percentile'` doubleCol)
+                                        )
+                                        (percentiles cfg)
+                                )
+                    SFalse ->
+                        map
+                            ((Col @(Maybe a) colName .==) . Lit . (`percentileOrd'` col))
+                            [1, 25, 75, 99]
             (UnboxedColumn (col :: VU.Vector a)) -> []
         columnConds = concatMap colConds [(l, r) | l <- columnNames df, r <- columnNames df]
           where
@@ -238,11 +292,9 @@ partitionDataFrame cond df = (filterWhere cond df, filterWhere (F.not cond) df)
 findBestSplit ::
     forall a.
     (Columnable a) =>
-    T.Text -> [Expr Bool] -> DataFrame -> Maybe (Expr Bool)
-findBestSplit target conds df =
+    TreeConfig -> T.Text -> [Expr Bool] -> DataFrame -> Maybe (Expr Bool)
+findBestSplit cfg target conds df =
     let
-        minLeafSize = 1
-        lambda = 0.05
         initialImpurity = calculateGini @a target df
         evalGain cond =
             let (t, f) = partitionDataFrame cond df
@@ -252,7 +304,8 @@ findBestSplit target conds df =
                 newImpurity =
                     (weightT * calculateGini @a target t)
                         + (weightF * calculateGini @a target f)
-             in ( (initialImpurity - newImpurity) - lambda * fromIntegral (eSize cond)
+             in ( (initialImpurity - newImpurity)
+                    - complexityPenalty (synthConfig cfg) * fromIntegral (eSize cond)
                 , negate (eSize cond)
                 )
 
@@ -262,7 +315,7 @@ findBestSplit target conds df =
                     let
                         (t, f) = partitionDataFrame c df
                      in
-                        nRows t >= minLeafSize && nRows f >= minLeafSize
+                        nRows t >= minLeafSize cfg && nRows f >= minLeafSize cfg
                 )
                 (nubOrd conds)
         sortedConditions = take 10 (sortBy (flip compare `on` evalGain) validConds)
@@ -273,7 +326,13 @@ findBestSplit target conds df =
                 Just $
                     maximumBy
                         (compare `on` evalGain)
-                        (boolExprs df sortedConditions sortedConditions 0 3)
+                        ( boolExprs
+                            df
+                            sortedConditions
+                            sortedConditions
+                            0
+                            (boolExpansion (synthConfig cfg))
+                        )
 
 calculateGini ::
     forall a.
